from __future__ import annotations

import functools
import inspect
from typing import TYPE_CHECKING, Any, AnyStr, Dict, List, Optional

import regex as re

from ruleau.constants import OverrideLevel
from ruleau.docs import clean_source, comments, description, doctests, parameters
from ruleau.exceptions import (
    CannotOverrideException,
    RuleIdIllegalCharacterException,
    RuleRequiresIdException,
    RuleRequiresNameException,
)
from ruleau.structures import DependentResults, ExecutionResult, RuleauDict

if TYPE_CHECKING:
    from ruleau.adapter import ApiAdapter
    from ruleau.process import Process
    from ruleau.types import Function


class Rule:
    def __init__(
        self,
        func: "Function",
        id_: AnyStr,
        name: AnyStr,
        depends_on: List["Rule"],
        override_level: OverrideLevel,
        lazy_dependencies: bool,
    ):
        """
        :param func: User defined rule
        :param name: User defined human readable name of the rule
        :param depends_on: Rule dependencies
        :param override_level: Override level
        :param lazy_dependencies: Flag to switch loading of rule dependencies lazily
        """
        self.id = id_
        self.name = name
        # Validate the rule, make sure the name is always set for a rule
        self.validate()
        # Set the user defined function
        self.func = func
        self.depends_on = depends_on
        self.override_level = override_level
        self.__name__ = func.__name__
        self.lazy_dependencies = lazy_dependencies
        self.order = None

        # This preserves the original Docstring on the decorated function
        # which allows DocTest to detect the function
        functools.update_wrapper(self, func)

    def __str__(self) -> str:
        return self.__name__

    def __call__(self, *args, **kwargs) -> bool:
        return self.func(*args, **kwargs)

    def validate(self):
        """
        Validator to check if top level rule has a human readable name and
        and id
        :raises: TopLevelRuleRequiresNameException
        :raises: RuleRequiresIdException
        """
        if not self.name or not isinstance(self.name, str):
            raise RuleRequiresNameException()
        if not self.id or not isinstance(self.id, str):
            raise RuleRequiresIdException()

        # Validate the Rule ID
        if not re.match(r"^([a-zA-Z0-9-_.~]+)+$", self.id):
            raise RuleIdIllegalCharacterException()

    def _get_source(self) -> AnyStr:
        return clean_source(inspect.getsource(self.func))

    @property
    def description(self) -> AnyStr:
        return description(self.func.__doc__)

    def calculate_order_values(self, index=None):
        """
        This sets the `order` value for this rule
        and all of the dependencies below it.

        :param index: Used internally, do not set
        """
        if index is None:
            self.order = 0
        else:
            if self.order is not None:
                self.order = max(self.order, index + 1)
            else:
                self.order = index + 1
        for rule in self.depends_on:
            rule.calculate_order_values(self.order)

    def flatten_rule_objects(self, flat_rules=None) -> [Rule]:
        if flat_rules is None:
            flat_rules = []
        flat_rules.append(self)
        for dependency in self.depends_on:
            dependency.flatten_rule_objects(flat_rules)
        return flat_rules

    def parse(self):
        return {
            "id": self.id,
            "name": self.name,
            "order": self.order,
            "override_level_name": self.override_level.name,
            "override_level": self.override_level.value,
            "source": self._get_source(),
            "comments": comments(self._get_source()),
            "docstring": self.func.__doc__,
            "description": self.description,
            "parameters": parameters(self.func.__doc__),
            "dependencies": [dependent.id for dependent in self.depends_on],
            "doctests": doctests(self.func),
        }

    def execute(
        self,
        case_id: AnyStr,
        payload: Dict[AnyStr, Any],
        process: "Process",
        api_adapter: Optional["ApiAdapter"] = None,
    ):
        api_result = {}
        # Create the rule result so that the execution can store the result
        if api_adapter:
            api_result = api_adapter.create_result(case_id, self, process)
        # Prep the rule payload
        rule_payload = RuleauDict(payload)
        # Prep the dependent results
        depend_results = DependentResults(
            case_id,
            self.depends_on,
            payload,
            process,
            api_adapter,
            lazy=self.lazy_dependencies,
        )
        # Prepare execution result for context from all dependencies
        context = ExecutionResult(self, process, rule_payload, None, depend_results)
        # Prepare execution result for the rule to be executed
        execution_result = ExecutionResult(
            self,
            process,
            rule_payload,
            self(context, rule_payload),
            depend_results,
        )
        # Store the rule result
        if api_adapter:
            # Apply overrides on the rule result
            execution_result = self.apply_override(
                case_id, execution_result, api_adapter
            )

            api_adapter.update_result(case_id, self, api_result, execution_result)
        # Return the rule result
        return execution_result

    def apply_override(
        self,
        case_id,
        execution_result: "ExecutionResult",
        api_adapter: ApiAdapter,
    ):
        # Get overrides for the rule in a case
        override = api_adapter.fetch_override(
            case_id, execution_result.process.id, self.id
        )

        # Apply override to the executed rule result, if any
        # Overrides should only be applied to allowed rule and if they're present
        if override:
            # Throw an exception if the backend is trying to override a NO_OVERRIDE rule
            if self.override_level == OverrideLevel.NO_OVERRIDE:
                raise CannotOverrideException(f"Cannot override {self.name}")
            else:
                # Override the rule result and set the overridden flag
                execution_result.override = override["id"]
                execution_result.original_result = execution_result.result
                execution_result.result = override["applied"]
        return execution_result



__version__ = "0.0.12"

from .adapter import ApiAdapter
from .aggregates import All, Any
from .constants import OverrideLevel
from .decorators import rule
from .execute import execute
from .structures import ExecutionResult, mock_context



from typing import TYPE_CHECKING, Any, AnyStr, Dict, Optional

from .rule import Rule

if TYPE_CHECKING:
    from ruleau.adapter import ApiAdapter


class Process:
    """Class holding rule process"""

    def __init__(
        self,
        process_id: str,
        name: str,
        description: str,
        root_rule: Rule,
    ):
        self.id = process_id
        self.name = name
        self.description = description
        self.root_rule = root_rule

    def parse(self):
        self.root_rule.calculate_order_values()
        flattened_rules = self.root_rule.flatten_rule_objects()
        unique_rules = list(set(flattened_rules))

        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "root_rule": self.root_rule.id,
            "rules": [rule.parse() for rule in unique_rules],
        }

    def execute(
        self,
        case_id: AnyStr,
        payload: Dict[AnyStr, Any],
        api_adapter: Optional["ApiAdapter"] = None,
    ):
        return self.root_rule.execute(case_id, payload, self, api_adapter)



import logging
from typing import Any, AnyStr, Dict, Optional
from urllib.parse import urljoin

import requests

from ruleau.decorators import api_request
from ruleau.exceptions import APIException
from ruleau.process import Process
from ruleau.rule import Rule

logger = logging.getLogger(__name__)


class ApiAdapter:
    base_url: AnyStr
    base_path: AnyStr
    api_key: Optional[AnyStr]

    def __init__(
        self,
        base_url: AnyStr,
        api_key: Optional[AnyStr] = None,
    ):
        """
        :param base_url: Base URL of the ruleau API
        :param api_key: (Optional) API key to authenticate with the API
        """
        self.base_url = base_url
        self.base_path = "/api/v1/"
        self.api_key = api_key

    @api_request
    def sync_case(self, case_id: AnyStr, process_id: AnyStr, payload: Dict) -> Dict:
        """
        Synchronise case with API
        :param case_id: The ID of the case being executed
        :param process_id: The ID of the process
        :param payload: Case payload to execute on
        :return:
        """
        response = requests.get(
            urljoin(
                self.base_url, f"{self.base_path}processes/{process_id}/cases/{case_id}"
            )
        )

        if response.status_code == 200:
            response = requests.patch(
                urljoin(
                    self.base_url,
                    f"{self.base_path}processes/{process_id}/cases/{case_id}",
                ),
                json={
                    "id": case_id,
                    "payload": payload,
                    "status": "OPEN",
                },
            )
            if response.status_code != 200:
                raise APIException(f"Failed to update case: {response.text}")

        elif response.status_code == 404:
            response = requests.post(
                urljoin(self.base_url, f"{self.base_path}processes/{process_id}/cases"),
                json={
                    "id": case_id,
                    "payload": payload,
                    "process": process_id,
                    "status": "OPEN",
                },
            )
            if response.status_code != 201:
                raise APIException(f"Failed to create case: {response.text}")

        else:
            raise APIException(f"Failed to check case: {response.text}")

        return response.json()

    @api_request
    def sync_process(self, process: Process):
        response = requests.post(
            urljoin(self.base_url, f"{self.base_path}processes"),
            json=process.parse(),
        )

        if response.status_code != 201:
            raise APIException(f"Unable to save rules: {response.text}")

        return response.json()

    @api_request
    def create_result(
        self,
        case_id: AnyStr,
        rule: Rule,
        process: Process,
    ) -> Dict:
        """
        Update rule result after its execution
        :param case_id: The ID of the case being executed
        :param rule: The Rule object that the result relates to
        :param process: The Process object that the rule is a child of
        :return:
        """
        # Update the rule results
        response = requests.post(
            urljoin(
                self.base_url,
                f"{self.base_path}processes/{process.id}/cases/"
                f"{case_id}/rules/{rule.id}/results",
            )
        )
        if response.status_code != 201:
            raise APIException(
                f"Failed to create result {case_id}@{rule.id}: {response.text}"
            )
        return response.json()

    @api_request
    def update_result(
        self,
        case_id: AnyStr,
        rule: Rule,
        api_result: Dict[AnyStr, Any],
        execution_result: "ExecutionResult",  # noqa: F821
    ) -> Dict:
        """
        Update rule result after its execution
        :param case_id:
        :param rule:
        :param api_result:
        :param execution_result:
        :return:
        """
        # Update the rule results
        response = requests.patch(
            urljoin(
                self.base_url,
                f"{self.base_path}processes/{execution_result.process.id}/cases/"
                f"{case_id}/rules/{rule.id}/results/{api_result['id']}",
            ),
            json={
                "result": execution_result.result,
                "payload": execution_result.payload.accessed,
                "override": execution_result.override,
                "original_result": execution_result.original_result,
            },
        )
        if response.status_code != 200:
            raise APIException(
                f"Failed to store rule result {case_id}@{rule.id}: {response.text}"
            )
        return response.json()

    @api_request
    def fetch_override(
        self, case_id: AnyStr, process_id: AnyStr, rule_id: AnyStr
    ) -> Optional[Dict[AnyStr, Any]]:
        """
        Fetch rule overrides
        :param case_id: client ID that identifies a previously established case
        :param process_id: The ID of the process that the case is being run against
        :param rule_id: The ID of the Rule to fetch overrides for
        :return: a ruleau overrides Optional[Dict[AnyStr, Any]]
        """
        response = requests.get(
            urljoin(
                self.base_url,
                f"{self.base_path}processes/{process_id}/"
                f"cases/{case_id}/overrides/search",
            ),
            params={"rule_id": rule_id},
        )
        if response.status_code != 200:
            return {}
        return response.json()



from ruleau import execute

from examples.kitchen_sink.rules import will_lend

if __name__ == "__main__":
    payload = {"data": {"fico_score": 150, "ccjs": [], "kyc": "low", "number_of_children": 1, "capital": 10_000}}
    result = execute(will_lend, payload, case_id="abc")



from typing import TYPE_CHECKING, AnyStr, Optional

from .dependent_results import MockDependentResults

if TYPE_CHECKING:
    from ruleau.process import Process
    from ruleau.rule import Rule
    from ruleau.structures import DependentResults, RuleauDict


class ExecutionResult:
    def __init__(
        self,
        executed_rule: "Rule",
        process: "Process",
        payload: "RuleauDict",
        result,
        dependent_results: "DependentResults",
        override: AnyStr = None,
        original_result: Optional[bool] = None,
    ):
        self.executed_rule = executed_rule
        self.process = process
        self.payload = payload
        self.result = result
        self.override = override
        self.original_result = original_result
        self.dependent_results = dependent_results


class MockExecutionResult(ExecutionResult):
    def __init__(self, dependent_results: MockDependentResults):
        self.dependent_results = dependent_results

    def parse(self):
        return {"dependent_results": self.dependent_results.parse()}


def mock_context(rule_results: dict[str, bool]):
    """
    Given an dictionary containing the rule function name
    as the key and the mocked result as the value, creates
    an ExecutionResult.

    :param rule_results: A dictionary containing the mock
    results for the depedent rules

    :return: ExecutionResult containing mocked results
    """

    dependent_results = MockDependentResults(rule_results)
    return MockExecutionResult(dependent_results)



#  flake8: noqa

from ruleau.structures.dependent_results import DependentResults
from ruleau.structures.execution_result import ExecutionResult, mock_context
from ruleau.structures.ruleau_dict import RuleauDict



from json.decoder import JSONDecodeError
from logging import getLogger
from typing import Optional

from requests.exceptions import RequestException

from ruleau.constants import OverrideLevel
from ruleau.exceptions import APIException
from ruleau.rule import Rule

logger = getLogger(__name__)
logger.propagate = True


def api_request(func):
    def _api_request(*args, **kwargs) -> Optional[dict]:
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.error(f"Exception occured: {e}")
            raise e

    return _api_request


def rule(
    rule_id,
    name,
    depends_on=None,
    override_level=OverrideLevel.ANY_OVERRIDE,
    lazy_dependencies=False,
):
    """Decorator to encapsulate a function into a rule"""
    depends_on = depends_on or []

    def _rule(func) -> Rule:
        return Rule(func, rule_id, name, depends_on, override_level, lazy_dependencies)

    return _rule



from examples.kitchen_sink.rules import will_lend
from ruleau import ApiAdapter, execute

if __name__ == "__main__":
    result = execute(
        will_lend,
        {"data": {"fico_score": 150, "ccjs": [], "kyc": "low", "number_of_children": 1, "capital": 10_000}},
        api_adapter=ApiAdapter(
            base_url="http://127.0.0.1:8081"
        ),  # Add API adapter to runner
        case_id="abc",
    )



class RuleRequiresNameException(Exception):
    """Exception raised if a rule doesn't have a human readable name"""


class RuleRequiresIdException(Exception):
    """Exception raised if rule doesn't have an id"""


class MethodNotAllowedException(Exception):
    """Exception raised if a forbidden RuleauDict method is called"""


class CaseIdRequiredException(Exception):
    """Exception raised if a json path for case identifier is not found"""


class CannotOverrideException(Exception):
    """Exception raised if a API tries to override a rule marked as NO_OVERRIDE"""


class DuplicateRuleNameException(Exception):
    """Exception raised if more than 1 rule has same name"""


class RuleIdIllegalCharacterException(Exception):
    """Exception raise if Rule name contains illegal characters"""


class DuplicateRuleIdException(Exception):
    """Exception raised if more than 1 rule has same name"""


class APIException(Exception):
    """Generic exception for API request failure"""



import argparse
import collections
import importlib
import json
import re
import sys
from argparse import Namespace
from doctest import DocTest, DocTestFinder, DocTestRunner, Example
from os import mkdir, path
from textwrap import dedent
from typing import TYPE_CHECKING, AnyStr, List, Optional

from jinja2 import Template
from jinja2.filters import FILTERS, environmentfilter

from ruleau.structures import mock_context

if TYPE_CHECKING:
    from ruleau.rule import Rule
    from ruleau.types import Function

PARAM_REGEX_PATTERN = ":(?P<name>[\*\s\w]+): (?P<description>.*?)\Z"  # noqa: W605


@environmentfilter
def json_print(_, value):
    """Pretty print json filter for jinja2"""
    value = value if isinstance(value, dict) or value is None else json.loads(value)
    return json.dumps(
        value,
        sort_keys=True,
        indent=4,
        separators=(",", ": "),
    )


#  Add the json_print filter to jinja2 filters
FILTERS["json_print"] = json_print


def trim(docstring):
    """trim function implementation from PEP-257
    https://www.python.org/dev/peps/pep-0257/#id18

    :param docstring: Docstring for a function
    """
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:  # pragma: no cover
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    if "\n" in docstring:
        trimmed.append("")
    # Return a single string:
    return "\n".join(trimmed)


def description(docstring: AnyStr) -> AnyStr:
    """Takes a rule docstring, and returns just the lines
    representing a description

    :param docstring: Docstring for a function
    """
    result = trim(docstring or "")
    lines = result.split("\n")
    end_of_description_idx = next(
        (
            idx
            for idx, line in enumerate(lines)
            if any(
                [
                    # Line is empty i.e. new line after description
                    line == "" or line.isspace(),
                    # Line is the start of a doc string parameter
                    re.match(PARAM_REGEX_PATTERN, line),
                    # Line is the start of a doc test
                    line.startswith(">>>"),
                ]
            )
        ),
        # If no end to the description then it ends with the string
        len(lines),
    )

    description_ = "\n".join(lines[:end_of_description_idx])
    # Remove final new line character if present
    return description_.removesuffix("\n")


def parameters(docstring: AnyStr):
    """Parse list of parameters out of the docstring

    :param docstring: The docstring of the rule
    :returns: [{"name": ..., "description": ...}, ...]
    """
    param_regex = re.compile(PARAM_REGEX_PATTERN)

    params = []
    if docstring:
        lines = docstring.split("\n")
        for line in lines:
            match = param_regex.findall(line.strip())
            if match:
                param_name, param_description = match[0]
                params.append({"name": param_name, "value": param_description})

    return params


def comments(source: AnyStr) -> AnyStr:
    """Returns single line comments from provided source

    :param source: clean source (output of clean_source)
    :returns: returns single line comments from source code
    """
    output_lines = [
        line for line in source.splitlines() if line.strip().startswith("#")
    ]
    return "\n".join(output_lines)


def get_doctest_rule_args(func: "Function", source: str, globs=None):
    """Given a rule function and the source, extract the arguments used when
    calling the rule in the given source.

    :param func: The rule function
    :param source: The source that calls the function
    :param globs: Globals for the rule execution
    """

    context = None
    payload = None

    #  Mock for rule function
    def mock_func(*args, **_):
        nonlocal context
        nonlocal payload
        #  We dont return early as the arguments for the final call of the
        # function should be used
        context, payload = args

    # Execute the funciton with the mocked rule
    globs = (globs or {}) | {func.__name__: mock_func, "mock_context": mock_context}
    exec(source, globs)

    # Return the context and payload used in the final call of the rule function
    return context, payload


def get_index_of_closing_bracket(source):
    """Returns index of closing bracket
    :param: source starting with opening bracket
    :returns: returns the index of the closing bracket that matches the opening
    bracket
    """
    # Create a deque to use it as a stack.
    queue = collections.deque()

    for i in range(len(source)):
        # Pop a starting bracket
        # for every closing bracket
        if source[i] == ")":
            queue.popleft()

        # Push all starting brackets
        elif source[i] == "(":
            queue.append(source[i])

        # If deque becomes empty
        if not queue:
            return i

    return None


def source_is_example(source, function_name):

    # Remove function name
    source = source.strip()
    if not source.startswith(function_name):
        return False
    source = source.removeprefix(function_name)
    source = source.strip()

    # Check after function name we find (
    if not source.startswith("("):
        return False

    # Find closing bracket of example
    closing_bracket_index = get_index_of_closing_bracket(source)
    if closing_bracket_index is None:
        return False

    #  If more happens after the call to the function
    if not source[closing_bracket_index:].strip() == ")":
        return False

    return True


def doctests(func: "Function") -> List[dict]:
    """Extract the doctest from a rule
    :param func: Function wrapped in a Rule instance
    :return: List of doctest examples
    """

    doctests = []
    finder = DocTestFinder(verbose=False, recurse=False)
    #  For each test
    for tests in finder.find(func):
        # Collect the source code up to the first result is expected
        source = ""
        for example in tests.examples:
            source = source + example.source

            # Once a result is expected run the source and extract the provided
            # values.
            if example.want.strip() != "":
                if source_is_example(example.source, func.__name__):
                    context, payload = get_doctest_rule_args(func, source, tests.globs)
                    doctests.append(
                        {
                            "test": source,
                            "result": example.want,
                            "context": json.dumps(context.parse()) if context else None,
                            "payload": json.dumps(payload),
                        }
                    )
                source = ""

    return doctests


def order_rules(rules: List["Rule"]) -> List["Rule"]:
    """Order rules so top level rules appear first"""
    is_root = {rule: True for rule in rules}

    # If any rule is dependent on a rule then it is not the root
    for rule in rules:
        for dependent in rules:
            if dependent != rule and dependent in rule.depends_on:
                is_root[dependent] = False

    roots = [rule for rule, is_root_rule in is_root.items() if is_root_rule]
    others = [rule for rule, is_root_rule in is_root.items() if not is_root_rule]
    return roots + others


def get_start_of_docstring(source_lines: List[AnyStr]) -> Optional[int]:
    """Get the index of line on which the doc string starts.

    :param source_lines: A list of the lines of the source of the function
    without the header or decorators
    """
    # Loop to find the start of the doc string
    for idx, line in enumerate(source_lines):

        # Skip any empty lines
        if line == "" or line.isspace():  # pragma: no cover
            # This is covered in the tests but no picked up by coverage
            continue

        # Find the start of the doc string
        if line.strip().startswith('"""'):
            return idx

        # If we find any non empty line that is not starting with """ then
        # there is no doc string for this function
        else:
            return None


def get_end_of_docstring(source_lines: List[AnyStr], start_idx=0) -> int:
    """Get the index of line on which the doc string ends.

    :param source_lines: A list of the lines of the source of the function
    without the header or decorators
    :param start_idx: The line on which the docstring starts

    :raise: Exception if no end of docstring found
    """
    for line_number, line in enumerate(source_lines[start_idx:]):  # pragma: no cover
        # A docstring cannot be started and not ended

        count = line.count('"""')  # Number of occurances of """ on line

        # If we are on the starting line (same as doc_string_start_idx) we
        # need to find 2 """ for the doc string to be closed.
        if (line_number == 0 and count == 2) or (line_number != 0 and count >= 1):
            return start_idx + line_number

    raise ValueError("No end of docstring found")


def clean_source(source: AnyStr) -> AnyStr:
    """Clean the source code.

    Remove all the lines until we have passed the decorators, function header
    and docstring.

    :param source: The source code to clean.
    """
    # Remove common indent
    source = dedent(source)

    # Get lines from source
    lines = source.split("\n")

    # Get line function header starts on. All function sources must have a header
    header_index = next(
        idx for idx, line in enumerate(lines) if line.startswith("def")
    )  # pragma: no cover

    # Remove anything above header from lines (decorators) and header it self
    lines = lines[header_index + 1 :]  # noqa: E203

    # Remove the docstring
    doc_string_start_idx = get_start_of_docstring(lines)

    # If we found a doc string
    if doc_string_start_idx is not None:
        doc_string_end_idx = get_end_of_docstring(lines, doc_string_start_idx)
        # Remove the doc string lines
        lines = lines[doc_string_end_idx + 1 :]  # noqa: E203

    clean_source_ = dedent("\n".join(lines))
    return clean_source_


def generate_documentation(rules) -> AnyStr:
    """Returns a HTML string, documenting the passed in rule and its dependents

    :param rules: List of rules to generate document of
    """
    with open(
        path.join(path.dirname(path.realpath(__file__)), "html", "documentation.html")
    ) as f:
        doc_template = Template(f.read())

    rules = order_rules(rules)
    enriched_rules = [rule.parse() for rule in rules]

    colors = {
        "light": "75FF96",
        "primary": "12C170",
        "dark": "17764A",
    }
    return doc_template.render(rules=enriched_rules, colors=colors)


def render_doc_for_module(module_file) -> AnyStr:
    """Render the documentation from a rule module
    :param module_file:
    :return:
    """
    spec = importlib.util.spec_from_file_location(module_file, module_file)
    config = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config)

    # Pick rules to document
    rules = [
        rule
        for rule in vars(config).values()
        if getattr(type(rule), "__name__", None) == "Rule"
    ]
    if not rules:
        raise Exception("No rules found in {}.".format(module_file))

    return generate_documentation(rules)


def generate_and_save_to_file(input_files, output_dir) -> int:
    """Generate the documentation and save to a file in provided output dir
    :param input_files: Rule modules provided by user via CLI
    :param output_dir: Output directory of generated HTML file
    :return:
    """
    generated_docs = []

    if len(input_files) == 0:
        raise Exception("No file(s) supplied to generate documentation for.")

    if not path.exists(output_dir):
        mkdir(output_dir)

    for target_file in input_files:
        if not path.exists(target_file):
            raise ValueError(f"{target_file} does not exist")

        documentation = render_doc_for_module(target_file)

        output_filename = path.join(
            output_dir, "{}.html".format(path.basename(target_file).split(".")[0])
        )

        with open(output_filename, "w") as f:
            f.write(documentation)
        generated_docs.append(output_filename)

    print(
        f"{len(generated_docs)} doc{'s' if len(generated_docs) > 1 else ''} generated."
    )
    print("\n".join(generated_docs))

    return 0


def get_arguments(args) -> Namespace:
    """Parses arguments for the ruleau-docs command
    :param args:
    :return:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default="./",
        help="Directory for generated documents",
    )
    parser.add_argument("files", nargs="*")
    return parser.parse_args(args)


def main() -> int:  # pragma: no cover
    """Console script for deft document generation.
    USAGE:
    ```bash
    $ ruleau-docs [--output-dir=<argument>] filename ...
    ```
    """
    args = get_arguments(sys.argv[1:])
    return generate_and_save_to_file(args.files, args.output_dir)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from copy import deepcopy
from json import dumps
from typing import TYPE_CHECKING, Any, AnyStr, Dict, Iterable, Optional

if TYPE_CHECKING:
    from ruleau.adapter import ApiAdapter
    from ruleau.process import Process
    from ruleau.rule import Rule


class DependentResults:
    def __init__(
        self,
        case_id: AnyStr,
        dependents: Iterable["Rule"],
        payload: Dict[AnyStr, Any],
        process: "Process",
        api_adapter: Optional["ApiAdapter"] = None,
        lazy: bool = False,
    ):
        self.case_id: AnyStr = case_id
        self.dependents = {dep.__name__: dep for dep in dependents}
        self.payload: Dict[AnyStr, Any] = deepcopy(payload)
        self.api_adapter = api_adapter
        self.results: Dict[AnyStr, Any] = {}
        self.process = process

        if not lazy:
            for depend in dependents:
                self.run(depend.__name__)

    def run(self, name):
        """
        Run and store the result of a rule dependency
        :param name:
        :return:
        """

        if name not in self.dependents:
            raise AttributeError(
                f"Result for rule '{name}' not available, as it was not "
                f"declared as a dependency. "
                f"depends_on={dumps(list(self.dependents.keys()))}"
            )
        # If the result of rule execution is not set, run & cache it
        if name not in self.results:
            self.results[name] = self.dependents[name].execute(
                self.case_id,
                self.payload,
                self.process,
                self.api_adapter,
            )
        # Return the rule execution result
        return self.results[name]

    def __getattr__(self, name):
        # Get the attribute otherwise, run the dependency
        return getattr(super(), name, self.run(name))

    def __iter__(self):
        # Iterate through the dependencies
        for dep in self.dependents:
            yield getattr(self, dep)


class MockDependentResults:
    def __init__(self, rule_results):
        self.rule_results = rule_results

    def __getattr__(self, name):
        return self.rule_results[name]

    def parse(self):
        return self.rule_results



import logging
from typing import TYPE_CHECKING, Any, AnyStr, Dict, List, Optional

from jsonpath_ng import parse

from ruleau.adapter import ApiAdapter
from ruleau.exceptions import (
    CaseIdRequiredException,
    DuplicateRuleIdException,
    DuplicateRuleNameException,
)
from ruleau.process import Process
from ruleau.rule import Rule

if TYPE_CHECKING:
    from ruleau.structures import ExecutionResult

logger = logging.getLogger(__name__)


def validate_no_duplicate_rule_names(rules: List[Rule]) -> None:
    """Returns True if there are no duplicate Rule Names are used
    A name can only be re-used if the same rule is included multiple times
    """
    rules_dict = {}
    for rule in rules:
        if rule.name not in rules_dict:
            rules_dict[rule.name] = rule
        else:
            if rule != rules_dict[rule.name]:
                raise DuplicateRuleNameException()


def validate_no_duplicate_rule_ids(rules: List[Rule]) -> None:
    """Returns True if there are no duplicate Rule IDs used
    An ID can only be re-used if the same rule is included multiple times
    """
    rules_dict = {}
    for rule in rules:
        if rule.id not in rules_dict:
            rules_dict[rule.id] = rule
        else:
            if rule != rules_dict[rule.id]:
                raise DuplicateRuleIdException()


def execute(
    executable_rule: Rule,
    payload: Dict[AnyStr, Any],
    case_id_jsonpath: AnyStr = None,
    case_id: Optional[AnyStr] = None,
    api_adapter: Optional[ApiAdapter] = None,
) -> "ExecutionResult":
    """
    Executes the provided rule, following dependencies and
    passing in results accordingly
    """

    # If neither case_id_jsonpath or case_id are present, raise exception
    if not case_id_jsonpath and not case_id:
        raise CaseIdRequiredException()

    # If case_id is not present in parameters, find it
    if not case_id:
        case_id_results = parse(case_id_jsonpath).find(payload)
        if not case_id_results:
            raise ValueError("Case ID not found in payload")
        case_id = str(case_id_results[0].value)

    # If there's no case ID, don't run
    if not case_id:
        raise ValueError("Case ID not found")

    # Validate unique rule name
    flattened_rules_as_objects = executable_rule.flatten_rule_objects()
    validate_no_duplicate_rule_names(flattened_rules_as_objects)

    # Validate unique rule ids
    validate_no_duplicate_rule_ids(flattened_rules_as_objects)

    # If API adapter was was passed sync the case
    executable_rule.calculate_order_values()
    process = Process(
        executable_rule.id,
        executable_rule.name,
        executable_rule.description,
        executable_rule,
    )

    if api_adapter:

        # Sync the process rules
        api_adapter.sync_process(process)

        # Sync the case
        api_adapter.sync_case(case_id, process.id, payload)

    # Trigger the rule execution, from the top level rule
    return process.execute(case_id, payload, api_adapter)
